print("import: 'vertica_python'")
import vertica_python

print("import: 'vertica_python.tests'")
import vertica_python.tests

print("import: 'vertica_python.vertica'")
import vertica_python.vertica

print("import: 'vertica_python.vertica.messages'")
import vertica_python.vertica.messages

print("import: 'vertica_python.vertica.messages.backend_messages'")
import vertica_python.vertica.messages.backend_messages

print("import: 'vertica_python.vertica.messages.frontend_messages'")
import vertica_python.vertica.messages.frontend_messages

